# -*- coding: utf-8 -*-
from imio.helpers.cache import invalidate_cachekey_volatile_for
from Products.PlonePAS.plugins.role import GroupAwareRoleManager
from Products.PluggableAuthService.plugins.ZODBRoleManager import ZODBRoleManager
from smtplib import SMTP_SSL
from zope.sendmail.mailer import SMTPMailer

import pkg_resources


try:
    pkg_resources.get_distribution("collective.solr")
except pkg_resources.DistributionNotFound:
    HAS_SOLR = False
else:
    from collective.solr.indexer import SolrIndexProcessor
    HAS_SOLR = True


def solr_index(self, obj, attributes=None):
    # Fix issue https://github.com/collective/collective.solr/issues/189
    if attributes is not None:
        attributes = None
    return SolrIndexProcessor._old_index(self, obj, attributes)


def ssl_makeMailer(self):
    """ Create a SMTPMailer with SMTP_SSL if port is 465 """
    mailer = SMTPMailer(hostname=self.smtp_host, port=int(self.smtp_port), username=self.smtp_uid or None,
                        password=self.smtp_pwd or None, force_tls=self.force_tls)
    if self.smtp_port in (465, '465'):
        mailer.smtp = SMTP_SSL
    return mailer


def assignRolesToPrincipal(self, roles, principal_id, REQUEST=None):  # noqa
    GroupAwareRoleManager._old_assignRolesToPrincipal(self, roles, principal_id, REQUEST)
    # we need to invalidate cachekey
    invalidate_cachekey_volatile_for('_users_groups_value')


def assignRoleToPrincipal(self, role_id, principal_id):
    ZODBRoleManager._old_assignRoleToPrincipal(self, role_id, principal_id)
    # we need to invalidate cachekey
    invalidate_cachekey_volatile_for('_users_groups_value')


def removeRoleFromPrincipal(self, role_id, principal_id):
    ZODBRoleManager._old_removeRoleFromPrincipal(self, role_id, principal_id)
    # we need to invalidate cachekey
    invalidate_cachekey_volatile_for('_users_groups_value')



# -*- coding: utf-8 -*-
from imio.helpers.content import get_vocab
from imio.helpers.testing import IntegrationTestCase
from plone import api
from zope.schema.tests.test_vocabulary import SimpleTermTests


class TestVocabularies(IntegrationTestCase):
    """
    Test vocabularies.
    """

    def test_SortedUsers(self):
        """ """
        api.user.create(email="test@test.org", username="a_new_user", password="secret")
        api.user.create(email="test@test.org", username="b_new_user", password="secret")
        api.user.create(email="test@test.org", username="new_user", password="secret")
        api.user.create(email="test@test.org", username="A_new_user", password="secret")
        api.user.create(email="test@test.org", username="B_new_user", password="secret")
        vocab = get_vocab(self.portal, "imio.helpers.SortedUsers")
        self.assertEqual(
            [term.title for term in vocab._terms],
            [u'a_new_user (a_new_user)',
             u'A_new_user (A_new_user)',
             u'b_new_user (b_new_user)',
             u'B_new_user (B_new_user)',
             u'new_user (new_user)',
             u'test_user_1_ (test_user_1_)'])
        vocab2 = get_vocab(self.portal, "imio.helpers.SimplySortedUsers")
        self.assertEqual(
            [term.title for term in vocab2._terms],
            [u'a_new_user',
             u'A_new_user',
             u'b_new_user',
             u'B_new_user',
             u'new_user',
             u'test_user_1_'])
        member = api.user.get(username="a_new_user")
        member.setMemberProperties({'fullname': 'anew User'})
        vocab = get_vocab(self.portal, "imio.helpers.SortedUsers")
        vocab2 = get_vocab(self.portal, "imio.helpers.SimplySortedUsers")
        self.assertEqual(
            [term.title for term in vocab._terms],
            [u'anew User (a_new_user)',
             u'A_new_user (A_new_user)',
             u'b_new_user (b_new_user)',
             u'B_new_user (B_new_user)',
             u'new_user (new_user)',
             u'test_user_1_ (test_user_1_)'])
        self.assertEqual(
            [term.title for term in vocab2._terms],
            [u'anew User',
             u'A_new_user',
             u'b_new_user',
             u'B_new_user',
             u'new_user',
             u'test_user_1_'])


class EnhancedTermTests(SimpleTermTests):
    """Inherited class tests are also called"""

    def _getTargetClass(self):
        from imio.helpers.vocabularies import EnhancedTerm
        return EnhancedTerm

    def test_attrs(self):
        term = self._makeOne(u'value', title=u'title')
        self.assertDictEqual(term.attrs, {})
        term = self._makeOne(u'value', title=u'title', attr1=u'attr1')
        self.assertDictEqual(term.attrs, {'attr1': u'attr1'})



# -*- coding: utf-8 -*-

import os
import re


def get_main_path(path='', subpath=''):
    """Return path/subpath if it exists. If path is empty, return buildout path."""
    if not path:
        # Are we in a classic buildout
        instance_home = os.getenv('INSTANCE_HOME')
        match = re.match('(.+)/parts/.+', instance_home)
        if match:
            path = match.group(1)
        else:
            path = os.getenv('PWD')
    if subpath:
        path = os.path.join(path, subpath)
    if os.path.exists(path):
        return path
    raise Exception("Path '{}' doesn't exist".format(path))


def relative_path(portal, fullpath):
    """Returns relative path following given portal (without leading slash).

    :param portal: leading object to remove from path
    :param fullpath: path to update
    :return: new path relative to portal object parameter
    """
    portal_path = '/'.join(portal.getPhysicalPath())  # not unicode, brain.getPath is also encoded
    if not fullpath.startswith(portal_path):
        return fullpath
    return fullpath[len(portal_path) + 1:]


def text_int_to_bool(item, key, log_method, log_params={}):
    """Changed to bool the value of item[key].

    :param item: yielded item (usually dict)
    :param key: dict key
    :param log_method: special log method (as log_error in imio.transmogrifier.iadocs.utils)
    :param log_params: log method parameters
    :return: the boolean value
    """
    try:
        return bool(int(item[key] or 0))
    except Exception:  # noqa
        log_method(item, u"Cannot change '{}' key value '{}' to bool".format(key, item[key]), **log_params)
    return False



# -*- coding: utf-8 -*-

from cgi import escape
from ZPublisher.Converters import type_converters

import json


# create type converter for :json
if 'json' not in type_converters:
    def field2json(v):
        try:
            v = json.loads(v)
        except ValueError:
            raise ValueError("Invalid json " + escape(repr(v)))
        return v
    type_converters['json'] = field2json



# -*- coding: utf-8 -*-
from imio.helpers.setup import load_type_from_package
from imio.helpers.setup import load_workflow_from_package
from imio.helpers.testing import IntegrationTestCase
from plone import api


class TestSetupModule(IntegrationTestCase):
    """
    Test all helper methods of setup module.
    """

    def setUp(self):
        self.portal = self.layer['portal']

    def test_load_workflow_from_package(self):
        wkf_tool = api.portal.get_tool('portal_workflow')
        wkf_obj = wkf_tool.get('intranet_workflow')
        states = wkf_obj.states
        self.assertIn('internal', states)
        states.deleteStates(['internal'])
        self.assertNotIn('internal', states)
        self.assertTrue(load_workflow_from_package(
            'intranet_workflow', 'profile-Products.CMFPlone:plone'))
        wkf_obj = wkf_tool.get('intranet_workflow')
        states = wkf_obj.states
        self.assertIn('internal', states)
        # not found WF
        self.assertFalse(load_workflow_from_package(
            'intranet_workflow2', 'profile-Products.CMFPlone:plone'))
        # not found profile_id
        self.assertFalse(load_workflow_from_package(
            'intranet_workflow', 'profile-Products.CMFPlone:plone2'))
        # WF not managed by given profile_id
        self.assertFalse(load_workflow_from_package(
            'comment_review_workflow', 'profile-Products.CMFPlone:plone'))

    def test_load_type_from_package(self):
        types_tool = api.portal.get_tool('portal_types')
        portal_type = types_tool.get('Folder')
        self.assertFalse(portal_type.filter_content_types)
        portal_type.filter_content_types = True
        self.assertTrue(portal_type.filter_content_types)
        self.assertTrue(load_type_from_package(
            'Folder', 'profile-Products.CMFPlone:plone'))
        portal_type = types_tool.get('Folder')
        self.assertFalse(portal_type.filter_content_types)
        # not found portal_type
        self.assertFalse(load_type_from_package(
            'Folder2', 'profile-Products.CMFPlone:plone'))
        # not found profile_id
        self.assertFalse(load_type_from_package(
            'Folder', 'profile-Products.CMFPlone:plone2'))
        # type not managed by given profile_id
        self.assertFalse(load_type_from_package(
            'testingtype', 'profile-Products.CMFPlone:plone'))



# -*- coding: utf-8 -*-

from DateTime import DateTime
from plone import api
from six.moves.cPickle import dumps
from time import time
from zope.globalrequest import getRequest
from zope.ramcache.ram import caches
from zope.ramcache.ram import RAMCache
from zope.ramcache.ram import Storage
from zope.ramcache.ram import writelock


class IMIORAMCache(RAMCache):
    """ """

    def _getStorage(self):
        """Finds or creates a storage object."""
        cacheId = self._cacheId
        with writelock:
            if cacheId not in caches:
                # imio.helpers, only changed here, use IMIOStorage instead Storage
                caches[cacheId] = IMIOStorage(
                    self.maxEntries, self.maxAge, self.cleanupInterval)
            return caches[cacheId]


class IMIOStorage(Storage):
    """ """

    def getEntry(self, ob, key):
        if self.lastCleanup <= time() - self.cleanupInterval:
            self.cleanup()

        try:
            data = self._data[ob][key]
        except KeyError:
            if ob not in self._misses:
                self._misses[ob] = 0
            self._misses[ob] += 1
            raise
        else:
            data[2] += 1                    # increment access count
            # XXX begin change by imio.helpers, update timestamp
            timestamp = time()
            data[1] = timestamp
            # XXX end change by PM

            return data[0]

    def getStatistics(self):
        objects = self._data.keys()
        objects.sort()
        result = []

        for ob in objects:
            try:
                size = len(dumps(self._data[ob]))
            except Exception as exc:
                size = 0
                api.portal.show_message(
                    'Could not compute size for "%s", original exception was "%s"'
                    % (repr(ob), repr(exc)), request=getRequest())
            hits = sum(entry[2] for entry in self._data[ob].itervalues())
            older_date = min(entry[1] for entry in self._data[ob].itervalues())
            result.append({'path': ob,
                           'hits': hits,
                           'misses': self._misses.get(ob, 0),
                           'size': size,
                           'entries': len(self._data[ob]),
                           'older_date': older_date and DateTime(older_date)})
        return tuple(result)


# instantiate our RAM Cache
imio_global_cache = IMIORAMCache()



# -*- coding: utf-8 -*-

from plone import api
from Products.CMFCore.WorkflowCore import WorkflowException
from Products.CMFPlone.utils import safe_unicode
from zope.i18n import translate

import logging


logger = logging.getLogger('imio.helpers.wokflow')


def do_transitions(obj, transitions, warn=False):
    """
        Apply multiple transitions on obj
    """
    wkf_tool = api.portal.get_tool("portal_workflow")
    if not isinstance(transitions, (list, tuple)):
        transitions = [transitions]
    for tr in transitions:
        try:
            wkf_tool.doActionFor(obj, tr)
        except WorkflowException as exc:
            if warn:
                logger.warn("Cannot apply transition '%s' on obj '%s': '%s'" % (tr, obj, exc))


def get_state_infos(obj):
    """ """
    wkf_tool = api.portal.get_tool('portal_workflow')
    review_state = wkf_tool.getInfoFor(obj, 'review_state')
    wf = wkf_tool.getWorkflowsFor(obj)[0]
    state = wf.states.get(review_state)
    state_title = state and state.title or review_state
    return {'state_name': review_state,
            'state_title': translate(safe_unicode(state_title),
                                     domain="plone",
                                     context=obj.REQUEST)}


def get_transitions(obj):
    """Return the ids of the available transitions as portal_workflow.getTransitionsFor
       will actually return a list of dict with various infos (id, title, name, ...) of
       the available transitions."""
    wf_tool = api.portal.get_tool('portal_workflow')
    return [tr["id"] for tr in wf_tool.getTransitionsFor(obj)]


def remove_state_transitions(wf_name, state_id, remv_trs=[]):
    """Removes transitions defined on a state and possible duplicates

    :param wf_name: workflow string name
    :param state_id: state id
    :param remv_trs: list of transitions to remove
    :return: True if removed
    """
    wf_tool = api.portal.get_tool('portal_workflow')
    # if wf_name not in wf_tool:  # prefer an exception...
    wf = wf_tool[wf_name]
    if state_id not in wf.states:
        return False
    state = wf.states[state_id]
    orig_len = len(state.transitions)
    trs = []
    # keep right transitions and remove duplicates
    [trs.append(tr) for tr in list(state.transitions) if tr not in remv_trs and tr not in trs]
    if len(trs) != orig_len:
        state.transitions = tuple(trs)
        return True
    return False



# -*- coding: utf-8 -*-

from imio.helpers.testing import IntegrationTestCase
from imio.helpers.transmogrifier import get_main_path
from imio.helpers.transmogrifier import relative_path
from imio.helpers.transmogrifier import text_int_to_bool

import os


class TestTesting(IntegrationTestCase):

    def test_get_main_path(self):
        orig_home = os.getenv('INSTANCE_HOME')
        orig_pwd = os.getenv('PWD')
        self.assertEquals(get_main_path('/etc'), '/etc')
        os.environ['INSTANCE_HOME'] = '/etc/parts/xx'
        self.assertEquals(get_main_path(), '/etc')
        os.environ['INSTANCE_HOME'] = '/etc/abcd'
        os.environ['PWD'] = '/myhome'
        self.assertRaisesRegexp(Exception, u"Path '/myhome' doesn't exist", get_main_path)  # path doesn't exist
        self.assertRaisesRegexp(Exception, u"Path '/myhome/toto' doesn't exist", get_main_path, subpath='toto')
        self.assertEquals(get_main_path('/', 'etc'), '/etc')
        os.environ['INSTANCE_HOME'] = orig_home
        os.environ['PWD'] = orig_pwd

    def test_relative_path(self):
        self.assertEquals(relative_path(self.portal, '/plone/directory'), 'directory')
        self.assertEquals(relative_path(self.portal, '/alone/directory'), '/alone/directory')

    def test_text_int_to_bool(self):
        dic = {'0': '0', 0: 0, '1': '1', 123: 123, 'a': 'a', 'errors': 0}

        def logger(item, msg):
            item['errors'] += 1

        self.assertFalse(text_int_to_bool(dic, '0', logger))  # using fake log method to test error
        self.assertFalse(text_int_to_bool(dic, 0, logger))
        self.assertTrue(text_int_to_bool(dic, '1', logger))
        self.assertTrue(text_int_to_bool(dic, 123, logger))
        self.assertEquals(dic['errors'], 0)
        self.assertFalse(text_int_to_bool(dic, 'a', logger))
        self.assertEquals(dic['errors'], 1)



# -*- coding: utf-8 -*-
from imio.helpers import get_cachekey_volatile
from imio.helpers.content import get_user_fullname
from natsort import humansorted
from operator import attrgetter
from plone import api
from plone.app.vocabularies.users import UsersFactory
from plone.memoize import ram
from zope.schema.vocabulary import SimpleTerm
from zope.schema.vocabulary import SimpleVocabulary


def voc_cache_key(method, self, context=None, query=''):
    """Returns a persistent portal stored date following the given cache key.

    Must be programatically invalidated."""
    return get_cachekey_volatile('_users_groups_value')


def get_users_voc(with_userid):
    acl_users = api.portal.get_tool('acl_users')
    users = acl_users.searchUsers(sort_by='')
    terms = []
    # manage duplicates, this can be the case when using LDAP and same userid in source_users
    userids = []
    for user in users:
        user_id = user['id']
        if user_id not in userids:
            userids.append(user_id)
            # bypass special characters, may happen when using LDAP
            try:
                unicode(user_id)
            except UnicodeDecodeError:
                continue
            if with_userid:
                term_title = u'{0} ({1})'.format(get_user_fullname(user_id), user_id)
            else:
                term_title = get_user_fullname(user_id)
            term = SimpleTerm(user_id, user_id, term_title)
            terms.append(term)
    terms = humansorted(terms, key=attrgetter('title'))
    return SimpleVocabulary(terms)


class SortedUsers(UsersFactory):
    """Create a users vocabulary with userid as value.
    Append ' (userid)' to term title."""

    @ram.cache(voc_cache_key)
    def SortedUsers__call__(self, context=None, query=''):
        return get_users_voc(True)

    __call__ = SortedUsers__call__


SortedUsersFactory = SortedUsers()


class SimplySortedUsers(SortedUsers):
    """With userid as value and fullname as title"""

    @ram.cache(voc_cache_key)
    def SimplySortedUsers__call__(self, context=None, query=''):
        return get_users_voc(False)

    __call__ = SimplySortedUsers__call__


SimplySortedUsersFactory = SimplySortedUsers()


class EnhancedTerm(SimpleTerm):

    def __init__(self, value, token=None, title=None, **attrs):
        super(EnhancedTerm, self).__init__(value, token=token, title=title)
        self.attrs = attrs



# -*- coding: utf-8 -*-
from imio.helpers.testing import IntegrationTestCase
from imio.helpers.workflow import do_transitions
from imio.helpers.workflow import get_state_infos
from imio.helpers.workflow import get_transitions
from imio.helpers.workflow import remove_state_transitions
from plone import api


class TestWorkflowModule(IntegrationTestCase):
    """
    Test all helper methods of workflow module.
    """

    def setUp(self):
        self.portal = self.layer['portal']

    def test_do_transitions(self):
        obj = api.content.create(container=self.portal.folder, id='mydoc', type='Document')
        self.assertEqual(api.content.get_state(obj), 'internal')
        # apply one transition
        do_transitions(obj, 'submit')
        self.assertEqual(api.content.get_state(obj), 'pending')
        # apply multiple transitions
        do_transitions(obj, ['publish_internally', 'publish_externally'])
        self.assertEqual(api.content.get_state(obj), 'external')
        do_transitions(obj, ['submit', 'retract', 'hide'])
        self.assertEqual(api.content.get_state(obj), 'private')

    def test_get_state_infos(self):
        self.portal.portal_workflow.setChainForPortalTypes(('Document', ), ('intranet_workflow', ))
        obj = api.content.create(container=self.portal.folder, id='mydoc', title='My document', type='Document')
        self.assertDictEqual(get_state_infos(obj), {'state_title': u'Internal draft', 'state_name': 'internal'})
        do_transitions(obj, 'submit')
        self.assertDictEqual(get_state_infos(obj), {'state_title': u'Pending review', 'state_name': 'pending'})

    def test_get_transitions(self):
        obj = api.content.create(container=self.portal.folder, id='mydoc', type='Document')
        self.assertEqual(get_transitions(obj),
                         ['publish_internally', 'hide', 'submit'])
        do_transitions(obj, 'submit')
        self.assertEqual(get_transitions(obj),
                         ['publish_internally', 'retract', 'publish_externally', 'reject'])
        do_transitions(obj, 'publish_externally')
        self.assertEqual(get_transitions(obj), ['retract'])

    def test_remove_state_transitions(self):
        wf_tool = api.portal.get_tool('portal_workflow')
        wf = wf_tool['intranet_workflow']
        self.assertFalse(remove_state_transitions('intranet_workflow', 'unknown_state'))
        self.assertTupleEqual(wf.states['internal'].transitions, ('hide', 'publish_internally', 'submit'))
        self.assertFalse(remove_state_transitions('intranet_workflow', 'internal'))  # nothing removed
        self.assertTrue(remove_state_transitions('intranet_workflow', 'internal', ['hide', 'submit']))  # removed
        self.assertTupleEqual(wf.states['internal'].transitions, ('publish_internally',))
        # add duplicates to simulate bad manipulation
        wf.states['internal'].transitions = tuple(['publish_internally', 'submit', 'submit'])
        self.assertTupleEqual(wf.states['internal'].transitions, ('publish_internally', 'submit', 'submit'))
        self.assertTrue(remove_state_transitions('intranet_workflow', 'internal'))
        self.assertTupleEqual(wf.states['internal'].transitions, ('publish_internally', 'submit'))
